"""Interface tests for Treants.

"""

import datreant as dtr
from datreant import Treant
import pytest
import mock
import os
import py

from . import test_collections
from .test_trees import TestTree


class TestTreant(TestTree):
    """Test generic Treant features"""
    treantname = 'testtreant'

    @pytest.fixture
    def treant(self, tmpdir):
        with tmpdir.as_cwd():
            c = dtr.treants.Treant(TestTreant.treantname)
        return c

    @pytest.fixture
    def basic_treant(self, tmpdir):
        # treant with tags and cats, in tmpdir
        with tmpdir.as_cwd():
            t1 = Treant('Rincewind')
            t1.tags.add('magical')
            t1.categories.add({'colour': 'octarine'})
            yield t1

    def test_init(self, treant, tmpdir):
        """Test basic Treant init"""
        assert treant.name == self.treantname
        assert treant.abspath == (os.path.join(tmpdir.strpath,
                                               self.treantname) + os.sep)

    def test_init_from_Tree(self, tmpdir):
        with tmpdir.as_cwd():
            tree = dtr.Tree('this')
            t = Treant(tree)

            assert t.path == tree.path

    @pytest.mark.parametrize("tags", (None, [], ['small', 'spiky']))
    @pytest.mark.parametrize("categories", (None, {}, {'colour': 'red'}))
    def test_init_generate(self, tags, categories, tmpdir):
        # test combinations of tags and categories
        # when generating from scratch
        with tmpdir.as_cwd():
            t = dtr.Treant('babs', tags=tags, categories=categories)

            if tags is not None:
                for tag in tags:
                    assert tag in t.tags
            if categories is not None:
                for cat, val in categories.items():
                    assert cat in t.categories
                    assert t.categories[cat] == val

    @pytest.mark.parametrize("tags", (None, [], ['small', 'spiky']))
    @pytest.mark.parametrize("categories", (None, {}, {'colour': 'red'}))
    def test_init_regenerate_via_name(self, tags, categories, tmpdir):
        # test regenerating a Treant from its directory
        with tmpdir.as_cwd():
            t = dtr.Treant('this')

            t2 = dtr.Treant('this', tags=tags, categories=categories)
            if tags is not None:
                for tag in tags:
                    assert tag in t2.tags
            if categories is not None:
                for cat, val in categories.items():
                    assert cat in t2.categories
                    assert t2.categories[cat] == val

    def test_gen_OSError(self, tmpdir):
        with tmpdir.as_cwd():
            with mock.patch('os.makedirs') as mp:
                mp.sideeffect = OSError(os.errno.ENOSPC, 'Mock - disk full')
                with pytest.raises(OSError) as error:
                    t = Treant('new')
                    t.tags.add('worthless')
                    assert error.errno == os.errno.ENOSPC

    def test_gen_OSError13(self, tmpdir):
        with tmpdir.as_cwd():
            with mock.patch('os.makedirs') as mp:
                mp.sideeffect = OSError(os.errno.EACCES, 'Mock - disk full')
                with pytest.raises(OSError) as error:
                    t = Treant('new')
                    t.tags.add('worthless')
                    assert error.errno == os.errno.EACCES
                    assert ("Permission denied; cannot create 'new'"
                            in str(error))

    def test_gen_methods(self, tmpdir):
        """Test the variety of ways we can generate a new Treant

        1. ``Treant('treant')``, where 'treant' is not an existing file or
           directory path

        2. ``Treant('treant')``, where 'treant' is an existing directory
           without Treant state files inside

        3. ``Treant('/somedir/treant')``, where 'treant' is not an existing
           file or directory in 'somedir'

        4. ``Treant('/somedir/treant')``, where 'treant' is an existing
           directory in 'somedir' without any Treant state files inside

        """
        with tmpdir.as_cwd():
            # 1
            t1 = Treant('newone')
            assert os.path.exists(t1._treantdir)

            # 2
            os.mkdir('another')
            t2 = Treant('another')
            assert os.path.exists(t2._treantdir)

            # 3
            t3 = Treant('yet/another')
            assert os.path.exists(t3._treantdir)

            # 4
            os.mkdir('yet/more')
            t4 = Treant('yet/more')
            assert os.path.exists(t4._treantdir)

    @pytest.mark.parametrize("tags", (None, [], ['small', 'spiky']))
    @pytest.mark.parametrize("categories", (None, {}, {'colour': 'red'}))
    def test_regen(self, tags, categories, tmpdir):
        """Test regenerating Treant.

        - create Treant
        - modify Treant a little
        - create same Treant (should regenerate)
        - check that modifications were saved
        """
        with tmpdir.as_cwd():
            C1 = Treant('regen', tags=tags, categories=categories)

            C2 = Treant('regen')  # should be regen of C1

            if tags is not None:
                for tag in tags:
                    assert tag in C2.tags
            if categories is not None:
                for cat, val in categories.items():
                    assert cat in C2.categories
                    assert C2.categories[cat] == val

            assert C1 == C2

            # they point to the same file, but they are not the same object
            assert C1 is not C2

    def test_cmp(self, tmpdir):
        """Test the comparison of Treants when sorting"""
        with tmpdir.as_cwd():
            c1 = Treant('a')
            c2 = Treant('b')
            c3 = Treant('c')

        assert sorted([c3, c2, c1]) == [c1, c2, c3]
        assert c1 <= c2 < c3
        assert c3 >= c2 > c1

    class TestTags:
        """Test treant tags"""

        @pytest.mark.parametrize("tags", (("1", "2", "3"),   # tuple
                                          ["1", "2", "3"],   # list
                                          {"1", "2", "3"}))  # set
        def test_adding_array_like_tags(self, treant, tags):
            treant.tags.add(tags)
            assert sorted(list(tags)) == sorted(list(treant.tags))

        def test_add_tags(self, treant):
            treant.tags.add('marklar')
            assert 'marklar' in treant.tags

            treant.tags.add('lark', 'bark')
            assert 'marklar' in treant.tags
            assert 'lark' in treant.tags
            assert 'bark' in treant.tags

        def test_remove_tags(self, treant):
            treant.tags.add('marklar')
            assert 'marklar' in treant.tags
            treant.tags.remove('marklar')
            assert 'marklar' not in treant.tags

            treant.tags.add('marklar')
            treant.tags.add('lark', 'bark')
            treant.tags.add(['fark', 'bark'])
            assert 'marklar' in treant.tags
            assert 'lark' in treant.tags
            assert 'bark' in treant.tags
            assert 'fark' in treant.tags
            assert len(treant.tags) == 4

            treant.tags.remove('fark')
            assert 'fark' not in treant.tags
            assert len(treant.tags) == 3
            treant.tags.remove('fark')
            assert len(treant.tags) == 3

            treant.tags.clear()
            assert len(treant.tags) == 0

        def test_tags_set_behavior(self, tmpdir):
            with tmpdir.as_cwd():
                # 1
                t1 = Treant('maple')
                t1.tags.add(['sprout', 'deciduous'])

                # 2
                t2 = Treant('sequoia')
                t2.tags.add(['sprout', 'evergreen'])

                tags_union = t1.tags | t2.tags
                for t in ['sprout', 'deciduous', 'evergreen']:
                    assert t in tags_union

                tags_intersect = t1.tags & t2.tags
                assert 'sprout' in tags_intersect
                for t in ['deciduous', 'evergreen']:
                    assert t not in tags_intersect

                tags_diff = t1.tags - t2.tags
                assert 'deciduous' in tags_diff
                for t in ['sprout', 'evergreen']:
                    assert t not in tags_diff

                tags_symm_diff = t1.tags ^ t2.tags
                for t in ['deciduous', 'evergreen']:
                    assert t in tags_symm_diff
                assert 'sprout' not in tags_symm_diff

                # 3
                t3 = Treant('oak')
                t3.tags.add(['deciduous'])

                # Test set membership
                assert t1.tags <= t1.tags
                assert not t1.tags < t1.tags
                assert t1.tags == t1.tags
                assert not t1.tags < t3.tags
                assert t1.tags > t3.tags

                # test TypeErrors in Tags
                # type_error_msg = "Operands must be AggTags, Tags, or a set."
                with pytest.raises(TypeError) as e:
                    ('tree') == t1.tags
                # assert e.value.message == type_error_msg
                with pytest.raises(TypeError) as e:
                    t1.tags < ('tree')
                # assert e.value.message == type_error_msg
                with pytest.raises(TypeError) as e:
                    ('tree') - t1.tags
                # assert e.value.message == type_error_msg
                with pytest.raises(TypeError) as e:
                    t1.tags - ('tree')
                # assert e.value.message == type_error_msg
                with pytest.raises(TypeError) as e:
                    ('tree') | t1.tags
                # assert e.value.message == type_error_msg
                with pytest.raises(TypeError) as e:
                    t1.tags | ('tree')
                # assert e.value.message == type_error_msg
                with pytest.raises(TypeError) as e:
                    ('tree') & t1.tags
                # assert e.value.message == type_error_msg
                with pytest.raises(TypeError) as e:
                    t1.tags & ('tree')
                # assert e.value.message == type_error_msg
                with pytest.raises(TypeError) as e:
                    ('tree') ^ t1.tags
                # assert e.value.message == type_error_msg
                with pytest.raises(TypeError) as e:
                    t1.tags ^ ('tree')
                # assert e.value.message == type_error_msg

        def test_tags_setting(self, tmpdir):
            """Test that we can set tags with lists or sets, or with Tags
            objects.

            """
            with tmpdir.as_cwd():

                # set with a list
                t1 = Treant('maple')
                t1.tags = ['sprout', 'deciduous']

                assert t1.tags == {'sprout', 'deciduous'}

                # set with a set
                t2 = Treant('elm')
                t2.tags = {'sprout', 'deciduous'}

                assert t2.tags == {'sprout', 'deciduous'}

                # set with a Tags object
                t3 = Treant('sequoia')
                t3.tags = t2.tags

                assert t3.tags == {'sprout', 'deciduous'}

        def test_tags_fuzzy(self, tmpdir, treant):
            """Test that fuzzy matching for tags works as expected.
            """
            treant.tags.add('bark', 'leafy', 'green', 'Leafy')

            for tag in ('leafy', 'Leafy'):
                assert tag in treant.tags.fuzzy('leafy')

        def test_tags_getitem(self, treant):
            """Test the queryability of tags via its __getitem__ method."""
            treant.tags.add('marklar', 'lark', 'bark')

            t = treant

            # single presence
            assert t.tags['lark']
            assert not t.tags['mark']

            # single not present
            assert t.tags[{'mark'}]

            # anding
            assert t.tags[['marklar', 'bark']]

            # oring
            assert t.tags['marklar', 'bark']
            assert t.tags['mark', 'bark']
            assert not t.tags['mark', 'dark']

            # not anding
            assert t.tags[{'dark', 'marklar'}]

            # complex logic
            assert t.tags[[('marklar', 'bark'), {'dark'}]]

        @pytest.mark.parametrize('tag', (1, 1.2))
        def test_tags_only_strings(self, treant, tag):
            with pytest.raises(ValueError):
                treant.tags.add(tag)

        def test_tags_printing(self, treant):
            treant.tags.add('marklar')
            repr = str(treant.tags)
            assert repr == "Tags\n====\n'marklar'\n"

    class TestCategories:
        """Test treant categories"""

        def test_add_categories(self, treant):
            treant.categories.add(marklar=42)
            assert 'marklar' in treant.categories

            treant.categories.add({'bark': 'snark'}, lark=27)
            assert 'bark' in treant.categories
            assert 'snark' not in treant.categories
            assert 'bark' in treant.categories

            assert treant.categories['bark'] == 'snark'
            assert treant.categories['lark'] == 27

            treant.categories['lark'] = 42
            assert treant.categories['lark'] == 42

        def test_remove_categories(self, treant):
            treant.categories.add(marklar=42)
            assert 'marklar' in treant.categories

            treant.categories.remove('marklar')
            assert 'marklar' not in treant.categories

            treant.categories.add({'bark': 'snark'}, lark=27)
            del treant.categories['bark']
            assert 'bark' not in treant.categories

            # should just work, even if key isn't present
            treant.categories.remove('smark')

            treant.categories['lark'] = 42
            treant.categories['fark'] = 32.3

            treant.categories.clear()
            assert len(treant.categories) == 0

        def test_add_wrong(self, treant):
            with pytest.raises(TypeError):
                treant.categories.add('temperature', 300)

            with pytest.raises(TypeError):
                treant.categories.add(['mark', 'matt'])

        @pytest.mark.parametrize('key, val', [[2, 'twenty'],
                                              [['blarg'], "nothin'"],
                                              [None, "literally nothin'"],
                                              [True, 'tautologically']])
        def test_add_wrong_keys(self, treant, key, val):
            # always test both addition methods
            with pytest.raises(TypeError):
                treant.categories[key] = val
            with pytest.raises(TypeError):
                treant.categories.add(key, val)

        @pytest.mark.parametrize('key, val', [['bark', ['shaggy']],
                                              ['snark', {'yes'}]])
        def test_add_wrong_values(self, treant, key, val):
            # always test both addition methods
            with pytest.raises(TypeError):
                treant.categories.add(key, val)
            with pytest.raises(TypeError):
                treant.categories[key] = val

        def test_None_deletes(self, treant):
            """Setting a category to ``None`` should delete the value.
            """
            treant.categories['size'] = 'large'
            treant.categories['bark'] = 'smooth'
            treant.categories['bark'] = None

            assert 'bark' not in treant.categories
            assert 'size' in treant.categories

            treant.categories['bark'] = 'smooth'
            treant.categories.add(bark=None)

            assert 'bark' not in treant.categories
            assert 'size' in treant.categories

        def test_setting_to_None_VE(self, treant):
            with pytest.raises(ValueError) as err:
                treant.categories['colour'] = None
            assert "Cannot set to 'None'" in str(err)

        def test_KeyError(self, treant):
            with pytest.raises(KeyError):
                treant.categories['hello?']

        def test_get_categories(self, treant):
            treant.categories['bark'] = 'dark'
            treant.categories['leaves'] = 'many'
            treant.categories['roots'] = 'shallow'

            # get a single category
            assert treant.categories['leaves'] == 'many'

            # get multiple categories with list
            assert treant.categories[['leaves', 'bark']] == ['many', 'dark']

            # get multiple categories with set
            assert treant.categories[{'leaves', 'bark'}] == {'leaves': 'many',
                                                             'bark': 'dark'}

        def test_set_categories(self, treant, tmpdir):
            a_dict = {'leaves': 'many', 'bark': 'dark'}

            treant.categories = a_dict
            assert treant.categories == a_dict

            a_dict.update({'roots': 'shallow'})

            assert treant.categories != a_dict
            treant.categories = a_dict
            assert a_dict == treant.categories

            # test setting from other Treant's categories
            with tmpdir.as_cwd():
                s = dtr.Treant('sprout')

            s.categories = {'shoes': False, 'shirt': False}

            assert treant.categories != s.categories

            treant.categories = s.categories
            assert treant.categories == s.categories

        def test_from_treant(self, treant, tmpdir):
            with tmpdir.as_cwd():
                dtr.Treant('sprout', categories=treant.categories)


class TestReadOnly:
    """Test Treant functionality when read-only"""

    @pytest.fixture
    def treant(self, tmpdir, request):
        with tmpdir.as_cwd():
            c = dtr.treants.Treant('testtreant')
            c.tags.add('72')
            py.path.local(c.abspath).chmod(0o0550, rec=True)

        def fin():
            py.path.local(c.abspath).chmod(0o0770, rec=True)

        request.addfinalizer(fin)

        return c

    def test_treant_read_only(self, treant):
        """Test that a read-only Treant can be accessed, but not written to.
        """
        c = dtr.treants.Treant(treant.abspath)

        assert '72' in c.tags

        with pytest.raises(OSError):
            c.tags.add('yet another')



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# datreant.tests

"""
datreant.tests --- unit tests for datreant
=============================================================

"""
# Bring some often used objects into the current namespace
from . import test_collections
from . import test_locks
from . import test_treants



# -*-encoding: utf-8 -*-
import os, sys
from functools import wraps

import numpy as np
import json
import datetime
try:
    import fcntl
except ImportError:
    sys.path.append(os.path.dirname(__file__))

from . import datreant as dt
from zipfile import *
from . import appdirs

data_path = appdirs.user_data_dir("open-moldyn")
tmp_path = data_path + "/tmp_sim"
tmp1_path = data_path + "/tmp_mdl"

CATEGORY_LIST = [
    "npart",
    "spc1",
    "spc2",
    "x_a",
    "timestep",
    "border_x_type",
    "border_y_type"
]

class ParamIO(dict):
    """
    An interface to interact with json files as a dictionary.

    Designed to be used with context manager (the with statement)
    and datreant.
    Upon entering the with statement it will attempt to load the json
    file at self.file_name into itself (as a dict).
    When leaving this context, it will store itself into the json file.

    Inherits dict.

    It will try to update the tags and categories of the treant dynState.

    Attributes
    ----------
    dynState : datreant.Treant
        The treant that support the leaf (file) associated with it.
    file_name : datreant.Leaf
        The file name of the json file associated with it.

    Example
    -------
    .. code-block:: python

        t = DynState(dirpath)
        # here t.open returns a ParamIO object as the file is .json
        with t.open("params.json") as IO:
            IO["my_key"] = "my_value"
            random_var = IO[some_key]
        # upon exiting the with block, the content of IO is stored
        # in the .json file
    """

    def __init__(self, dynState : dt.Treant, file: dt.Leaf, **kwargs):
        """

        Parameters
        ----------
        dynState
        file
        kwargs
        """
        super().__init__(**kwargs)
        self.dynState = dynState
        self.file_name = file

    def __enter__(self):
        """
        Try to load the content of the json file into itself

        Try to open the json file from file_name and load the informtions
        it contains into itself.
        Will catch FileNotFoundError and JSONDecodeError

        Returns
        -------
        self : ParamIO
        """
        try:
            with open(self.file_name, mode='r') as file:
                params = json.load(self.file_name)
            for key, value in params.items():
                self[key] = value
        except json.decoder.JSONDecodeError:
            print("File corrupted")
            pass
        except FileNotFoundError:
            #print("File does not YET exists")
            pass

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Open the json file file_name and dump itself into it.

        Open the json file file_name and dump itself into it +
        update the tags and categories of dynState according to
        the CATEGORY_LIST of the module.

        Parameters
        ----------
        exc_type
        exc_val
        exc_tb
        """
        param_exists = False
        try:
            with open(self.file_name, mode='r') as file:
                params = json.load(self.file_name)
                param_exists = True
        except json.decoder.JSONDecodeError:
            print("File corrupted")
            pass
        except FileNotFoundError:
            #print("File does not YET exists")
            pass
        if not param_exists or self != params:
            with open(self.file_name, mode='w') as file:
                json.dump(self, file, ensure_ascii=False, indent=4)
            #self.dynState.categories['last modified'] = datetime.datetime.now().strftime('%d/%m/%Y-%X')
        #self._update_categories()

    def from_dict(self, rdict: dict):
        """
        Copy rdict into itself

        Parameters
        ----------
        rdict : dict
            The remote dictionary from which to copy

        Returns
        -------

        """
        for key, value in rdict.items():
            self[key] = value

    def to_dict(self, rdict : dict):
        """
        Copy itself into the remote dictionary
        Parameters
        ----------
        rdict : dict
            The remot dictionary to which to copy

        Returns
        -------

        """
        for key, value in self.items():
            rdict[key] = value

    def to_attr(self, obj):
        """
        Dump the parameter dictionary in the object (obj) as attributes of said object.

        Warning
        -------
        Will change the value of each attribute of obj that have a name
        corresponding to a key of ParamIO.

        Parameters
        ----------
        obj : an object
            The object to which it will dump its content as attributes

        """
        for key, value in self.items():
            obj.__setattr__(str(key), value)

    def _update_categories(self):
        if self["x_a"] != 0:
            self.dynState.categories["nb_species"] = 2
        for key, value in self.items():
            if key in CATEGORY_LIST:
                self.dynState.categories[key] = value

class NumpyIO:
    """
    An interface to interact with numpy save files with a context manager.

    Attributes
    ----------
    dynState : datreant.Treant
        The treant that support the leaf (file) associated with it.
    mode : str
        The access mode of the file (usually 'r' or 'w' or 'a')
    file_name : datreant.Leaf
        The file name of the .npy file associated with it.
    file : file object
        the file that is opened (contains None until entering a context manager)

    Example
    -------
    .. code-block:: python

        t = DynState(dirpath)
        # here t.open returns a NumpyIO object as the file is .npy
        with t.open("pos.npy", 'r') as IO:
            arr = IO.load() #load an array
        with t.open("pos.npy", 'w') as IO:
            IO.save(arr) #save an array
    """
    def __init__(self, dynState, file : dt.Leaf, mode):
        self.dynState = dynState
        self.mode = mode
        self.file_name = file
        self.file = None

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def open(self):
        """
        Open the internal file.
        """
        self.file = open(self.file_name, mode=self.mode)

    def close(self):
        """
        Close the internal file.
        """
        self.file.close()

    def save(self, arr):
        """
        Save an array to the opened file.

        Parameters
        ----------
        arr : ndarray (numpy)
            The array to be stored
        Returns
        -------

        """
        np.save(self.file, arr)

    def load(self):
        """
        Load an array stored in the file.

        Returns
        -------
        arr : ndarray
            the array loaded
        """
        return np.load(self.file, allow_pickle=True)


class DynState(dt.Treant):
    """
    A Treant specialized for handling .npy and .json files for moldyn

    Attributes
    ----------
    POS: str
        standard name of the position file ("pos.npy")
    POS_H: str
        standard name of the position history file ("pos_history.npy")
    VEL: str
        standard name of the velocity file ("velocities.npy")
    STATE_FCT: str
        standard name of the state function file ("state_fct.json")
    PAR: str
        standard name of the parameter file ("parameters.json")
    """
    POS = "pos.npy"  # position of particles
    POS_H = "pos_history.npy"  # history of position
    VEL = "velocities.npy"  # final velocities
    STATE_FCT = "state_fct.json"  # state functions (energy, temperature...)
    PAR = "parameters.json"  # parameters of model and simulation

    def __init__(self, treant, *, extraction_path : str = data_path+'/tmp'):
        if isinstance(treant, dt.Treant):
            super().__init__(treant)
        elif isinstance(treant, str) and is_zipfile(treant) :
            try:
                with ZipFile(treant, 'r') as archive:
                    archive.extractall(extraction_path)
                    super().__init__(extraction_path)
            except BadZipFile:
                pass
        else:
            super().__init__(treant)

    def open(self, file, mode='r'):
        """
        Open the file in this tree (ie. directory and subdir).
        Return the appropriate IO class depending of the type of the file.

        If the type is not recognize, it opens the file and return the
        BytesIO or StringIO object.

        Parameters
        ----------
        file: str
            The name of the file. This file must be in this tree (ie. directory and subdir).
        mode: str (default='r')
            The mode with which to open the file :
                - 'r' to read
                - 'w' to write
                - 'a' to append
                - 'b' to read or write bytes (eg. 'w+b' to write bytes)

        Returns
        -------
        If file is a .npy file, return a NumpyIO object.
        If file is a .json file, return a ParamIO object.
        Else, return a StringIO or a BytesIO depending on mode.

        Note
        ----

        This method is designed to be used wih a context manager like this

        .. code-block:: python

            t = DynState(dirpath)
            # here t.open returns a NumpyIO object as the file is .npy
            with t.open("pos.npy", 'r') as IO:
                arr = IO.load() #load an array
            with t.open("pos.npy", 'w') as IO:
                IO.save(arr) #save an array
        """
        if file.endswith(".npy"):
            if not(mode.endswith("+b")):
                mode += "+b"
            return NumpyIO(self, self.leafloc[file], mode)
        elif file.endswith(".json"):
            return ParamIO(self, self.leafloc[file])
        else:
            return open(self.leafloc[file].abspath, mode)

    def add_tag(self,*tags):
        self.tags.add(*tags)

    def to_zip(self, path: str):
        """
        Zip every leaf (aka. file) of the dynState treant into an archive at path.

        Parameters
        ----------
        path : str
            The path of the archive.
        """
        with ZipFile(path, "w") as archive:
            for leaf in self.leaves():
                if leaf.exists:
                    leaf_path = str(leaf.abspath).replace('\\', '/').split('/')[-1]
                    archive.write(leaf.relpath, arcname=leaf_path)

    def save_model(self, model):
        """
        Save the positions, the velocities and the parameters of the model.

        The position and velocity arrays are saved as numpy files and
        the parameter dictionary as a .json file

        Parameters
        ----------
        model : simulation.builder.Model
            The model to be saved.
        """

        # position of particles
        with self.open(self.POS, 'w') as IO:
            IO.save(model.pos)
        # parameters
        with self.open(self.PAR, 'w') as IO:
            IO.from_dict(model.params)
        # velocity
        with self.open(self.VEL, 'w') as IO:
            IO.save(model.v)

@wraps(dt.discover)
def discover(dirpath=data_path, *args, **kwargs):
    return dt.discover(dirpath=dirpath, *args, **kwargs)



